import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:osam/domain/event/event.dart';
import 'package:osam/domain/middleware/middleware.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/domain/usecase/usecase.dart';

typedef ErrorHandler = void Function(Object error);

class UseCaseNotFoundException implements Exception {
  final Type useCaseType;

  UseCaseNotFoundException(this.useCaseType);
}

class Store<ST extends BaseState<ST>> {
  ST state;
  final List<Middleware<ST>> middleWares;
  final List<UseCase<ST>> useCases;
  final ErrorHandler errorHandler;

  var logging = false;

  Stream<Event<ST>> get eventStream => _dispatcher.stream;

  U useCase<U extends UseCase<ST>>() {
    final useCase = _useCasesMap[U];
    if (useCase == null) {
      throw UseCaseNotFoundException(U);
    } else
      return useCase;
  }

  // ignore: close_sinks
  final _dispatcher = StreamController<Event<ST>>.broadcast();
  final _denormalizedConditions = <Condition<ST>>[];
  final _useCasesMap = <Type, UseCase<ST>>{};

  Store(this.state, {this.middleWares, this.useCases, this.errorHandler}) {
    if (useCases.isNotEmpty) {
      for (final useCase in useCases) {
        useCase.store = this;
        _useCasesMap[useCase.runtimeType] = useCase;
      }
    }
    if (middleWares.isNotEmpty) {
      _setStoreToMiddlewares(middleWares);
      _denormalizeConditions();
    }
    _dispatcher.stream.listen((event) {
      try {
        if (logging) print(event.toString());

        for (final condition in _denormalizedConditions) {
          condition(event);
        }

        event(state);
      } catch (e) {
        dispatchEvent(event: ErrorEvent(e));
        errorHandler?.call(e);
      }
    });
  }

  void dispatchEvent({@required Event<ST> event}) => _dispatcher.sink.add(event);

  void _denormalizeConditions() {
    _denormalizedConditions.clear();
    _denormalizedConditions
        .addAll(middleWares.expand((middleWare) => middleWare.conditions).toList());
  }

  void _setStoreToMiddlewares(List<Middleware<ST>> middlewares) =>
      middlewares.forEach((middleware) => middleware.store = this);

  void updateMiddlewares(List<Middleware<ST>> middlewares) {
    _setStoreToMiddlewares(middlewares);
    middlewares.forEach((middleware) {
      if (middleWares.contains(middleware)) {
        middleWares[middleWares.indexOf(middleware)] = middleware;
      } else {
        middleWares.add(middleware);
      }
    });
    _denormalizeConditions();
  }
}
